/*
 * Copyright (C) 1997-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file client_common.c Common functions of para_client and para_audiod. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "crypt.h"
#include "net.h"
#include "fd.h"
#include "string.h"
#include "client.cmdline.h"
#include "client.h"
#include "buffer_tree.h"
#include "version.h"

/** The size of the receiving buffer. */
#define CLIENT_BUFSIZE 4000

/**
 * Close the connection to para_server and deallocate per-command ressources.
 *
 * \param ct The client task.
 *
 * This frees all ressources of the current command but keeps the configuration
 * in \p ct->conf.
 *
 * \sa \ref client_close().
 */
void client_disconnect(struct client_task *ct)
{
	if (!ct)
		return;
	if (ct->scc.fd >= 0)
		close(ct->scc.fd);
	sc_free(ct->scc.recv);
	ct->scc.recv = NULL;
	sc_free(ct->scc.send);
	ct->scc.send = NULL;
	btr_free_node(ct->btrn);
	ct->btrn = NULL;
}

/**
 * Close the connection to para_server and free all resources.
 *
 * \param ct Pointer to the client data.
 *
 * \sa \ref client_open(), \ref client_disconnect().
 */
void client_close(struct client_task *ct)
{
	if (!ct)
		return;
	client_disconnect(ct);
	free(ct->user);
	free(ct->config_file);
	free(ct->key_file);
	client_cmdline_parser_free(&ct->conf);
	free(ct);
}

/**
 * The preselect hook for server commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * The task pointer must contain a pointer to the initialized client data
 * structure as it is returned by client_open().
 *
 * This function checks the state of the connection and adds the file descriptor
 * of the connection to the read or write fd set of \a s accordingly.
 *
 * \sa register_task() client_open(), struct sched, struct task.
 */
static void client_pre_select(struct sched *s, struct task *t)
{
	int ret;
	struct client_task *ct = container_of(t, struct client_task, task);
	struct btr_node *btrn = ct->btrn;

	if (ct->scc.fd < 0)
		return;
	switch (ct->status) {
	case CL_CONNECTED:
	case CL_SENT_AUTH:
	case CL_SENT_CH_RESPONSE:
	case CL_SENT_COMMAND:
		para_fd_set(ct->scc.fd, &s->rfds, &s->max_fileno);
		return;

	case CL_RECEIVED_WELCOME:
	case CL_RECEIVED_PROCEED:
		para_fd_set(ct->scc.fd, &s->wfds, &s->max_fileno);
		return;

	case CL_RECEIVING:
		ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
		if (ret != 0) {
			if (ret < 0)
				sched_min_delay(s);
			else
				para_fd_set(ct->scc.fd, &s->rfds,
					&s->max_fileno);
		}
		return;
	case CL_SENDING:
		ret = btr_node_status(btrn, 0, BTR_NT_LEAF);
		if (ret != 0) {
			if (ret < 0)
				sched_min_delay(s);
			else
				para_fd_set(ct->scc.fd, &s->wfds,
					&s->max_fileno);
		}
		return;
	}
}

static int client_recv_buffer(struct client_task *ct, fd_set *rfds,
		char *buf, size_t sz, size_t *n)
{
	int ret;

	if (ct->status < CL_SENT_CH_RESPONSE)
		return read_nonblock(ct->scc.fd, buf, sz, rfds, n);

	*n = 0;
	ret = sc_recv_buffer(&ct->scc, buf, sz);
	/*
	 * sc_recv_buffer is used with blocking fds elsewhere, so it
	 * does not use the nonblock-API. Therefore we need to
	 * check for EOF and EAGAIN.
	 */
	if (ret == 0)
		return -E_SERVER_EOF;
	if (ret == -ERRNO_TO_PARA_ERROR(EAGAIN))
		return 0;
	if (ret < 0)
		return ret;
	*n = ret;
	return 0;
}

/**
 * The post select hook for client commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * Depending on the current state of the connection and the status of the read
 * and write fd sets of \a s, this function performs the necessary steps to
 * authenticate the connection, to send the command given by \a t->private_data
 * and to receive para_server's output, if any.
 *
 * \sa struct sched, struct task.
 */
static void client_post_select(struct sched *s, struct task *t)
{
	struct client_task *ct = container_of(t, struct client_task, task);
	struct btr_node *btrn = ct->btrn;
	int ret = 0;
	size_t n;
	char buf[CLIENT_BUFSIZE];

	t->error = 0;
	if (ct->scc.fd < 0)
		return;
	switch (ct->status) {
	case CL_CONNECTED: /* receive welcome message */
		ret = client_recv_buffer(ct, &s->rfds, buf, sizeof(buf), &n);
		if (ret < 0 || n == 0)
			goto out;
		ct->status = CL_RECEIVED_WELCOME;
		return;
	case CL_RECEIVED_WELCOME: /* send auth command */
		sprintf(buf, AUTH_REQUEST_MSG "%s", ct->user);
		PARA_INFO_LOG("--> %s\n", buf);
		if (!FD_ISSET(ct->scc.fd, &s->wfds))
			return;
		ret = send_buffer(ct->scc.fd, buf);
		if (ret < 0)
			goto out;
		ct->status = CL_SENT_AUTH;
		return;
	case CL_SENT_AUTH:
		/*
		 * Receive challenge and session keys, decrypt the challenge and
		 * send back the hash of the decrypted challenge.
		 */
		{
		/* decrypted challenge/session key buffer */
		unsigned char crypt_buf[1024];
		/* the SHA1 of the decrypted challenge */
		unsigned char challenge_hash[HASH_SIZE];

		ret = client_recv_buffer(ct, &s->rfds, buf, sizeof(buf), &n);
		if (ret < 0 || n == 0)
			goto out;
		PARA_INFO_LOG("<-- [challenge] (%zu bytes)\n", n);
		ret = priv_decrypt(ct->key_file, crypt_buf,
			(unsigned char *)buf, n);
		if (ret < 0)
			goto out;
		hash_function((char *)crypt_buf, CHALLENGE_SIZE, challenge_hash);
		ct->scc.send = sc_new(crypt_buf + CHALLENGE_SIZE, SESSION_KEY_LEN);
		ct->scc.recv = sc_new(crypt_buf + CHALLENGE_SIZE + SESSION_KEY_LEN,
			SESSION_KEY_LEN);
		hash_to_asc(challenge_hash, buf);
		PARA_INFO_LOG("--> %s\n", buf);
		ret = send_bin_buffer(ct->scc.fd, (char *)challenge_hash,
			HASH_SIZE);
		if (ret < 0)
			goto out;
		ct->status = CL_SENT_CH_RESPONSE;
		return;
		}
	case CL_SENT_CH_RESPONSE: /* read server response */
		{
		ret = client_recv_buffer(ct, &s->rfds, buf, sizeof(buf), &n);
		if (ret < 0 || n == 0)
			goto out;
		/* check if server has sent "Proceed" message */
		ret = -E_CLIENT_AUTH;
		if (n < PROCEED_MSG_LEN)
			goto out;
		if (!strstr(buf, PROCEED_MSG))
			goto out;
		ct->status = CL_RECEIVED_PROCEED;
		return;
		}
	case CL_RECEIVED_PROCEED: /* concat args and send command */
		{
		int i;
		char *command = NULL;
		if (!FD_ISSET(ct->scc.fd, &s->wfds))
			return;
		for (i = 0; i < ct->conf.inputs_num; i++) {
			char *tmp = command;
			command = make_message("%s\n%s", command?
				command : "", ct->conf.inputs[i]);
			free(tmp);
		}
		command = para_strcat(command, EOC_MSG "\n");
		PARA_DEBUG_LOG("--> %s\n", command);
		ret = sc_send_buffer(&ct->scc, command);
		free(command);
		if (ret < 0)
			goto out;
		ct->status = CL_SENT_COMMAND;
		return;
		}
	case CL_SENT_COMMAND:
		{
		char *buf2;
		/* can not use "buf" here because we need a malloced buffer */
		buf2 = para_malloc(CLIENT_BUFSIZE);
		ret = client_recv_buffer(ct, &s->rfds, buf2, CLIENT_BUFSIZE, &n);
		if (n > 0) {
			if (strstr(buf2, AWAITING_DATA_MSG)) {
				free(buf2);
				ct->status = CL_SENDING;
				return;
			}
			ct->status = CL_RECEIVING;
			btr_add_output(buf2, n, btrn);
		} else
			free(buf2);
		goto out;
		}
	case CL_SENDING:
		{
		char *buf2;
		size_t sz;
		ret = btr_node_status(btrn, 0, BTR_NT_LEAF);
		if (ret < 0)
			goto out;
		if (ret == 0)
			return;
		if (!FD_ISSET(ct->scc.fd, &s->wfds))
			return;
		sz = btr_next_buffer(btrn, &buf2);
		ret = sc_send_bin_buffer(&ct->scc, buf2, sz);
		if (ret < 0)
			goto out;
		btr_consume(btrn, sz);
		return;
		}
	case CL_RECEIVING:
		{
		char *buf2;
		ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
		if (ret < 0)
			goto out;
		if (ret == 0)
			return;
		/*
		 * The FD_ISSET() is not strictly necessary, but is allows us
		 * to skip the malloc below if there is nothing to read anyway.
		 */
		if (!FD_ISSET(ct->scc.fd, &s->rfds))
			return;
		buf2 = para_malloc(CLIENT_BUFSIZE);
		ret = client_recv_buffer(ct, &s->rfds, buf2, CLIENT_BUFSIZE, &n);
		if (n > 0) {
			buf2 = para_realloc(buf2, n);
			btr_add_output(buf2, n, btrn);
		} else
			free(buf2);
		goto out;
		}
	}
out:
	t->error = ret;
	if (ret < 0) {
		if (ret != -E_SERVER_EOF && ret != -E_BTR_EOF)
			PARA_ERROR_LOG("%s\n", para_strerror(-t->error));
		btr_remove_node(btrn);
	}
}

/**
 * Connect to para_server and register the client task.
 *
 * \param ct The initialized client task structure.
 * \param s The scheduler instance to register the client task to.
 * \param parent The parent node of the client btr node.
 * \param child The child node of the client node.
 *
 * The client task structure given by \a ct  must be allocated and initialized
 * by \ref client_parse_config() before this function is called.
 *
 * \return Standard.
 */
int client_connect(struct client_task *ct, struct sched *s,
		struct btr_node *parent, struct btr_node *child)
{
	int ret;

	PARA_NOTICE_LOG("connecting %s:%d\n", ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	ct->scc.fd = -1;
	ret = para_connect_simple(IPPROTO_TCP, ct->conf.hostname_arg,
					       ct->conf.server_port_arg);
	if (ret < 0)
		return ret;
	ct->scc.fd = ret;
	ret = mark_fd_nonblocking(ct->scc.fd);
	if (ret < 0)
		goto err_out;
	ct->status = CL_CONNECTED;
	ct->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "client", .parent = parent, .child = child));
	ct->task.pre_select = client_pre_select;
	ct->task.post_select = client_post_select;
	ct->task.error = 0;
	sprintf(ct->task.status, "client");
	register_task(s, &ct->task);
	return 1;
err_out:
	close(ct->scc.fd);
	ct->scc.fd = -1;
	return ret;
}

/**
 * Parse a client configuration.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 * \param ct_ptr Filled in by this function.
 * \param loglevel If not \p NULL, the number of the loglevel is stored here.
 *
 * This checks the command line options given by \a argc and \a argv, sets
 * default values for the user name and the name of the rsa key file and reads
 * further options from the config file.
 *
 * Upon successful return, \a ct_ptr points to a dynamically allocated and
 * initialized client task struct.
 *
 * \return The number of non-option arguments in \a argc/argv on success,
 * negative on errors.
 */
int client_parse_config(int argc, char *argv[], struct client_task **ct_ptr,
		int *loglevel)
{
	char *home = para_homedir();
	int ret;
	struct client_task *ct = para_calloc(sizeof(struct client_task));

	*ct_ptr = ct;
	ct->scc.fd = -1;
	ret = -E_CLIENT_SYNTAX;
	if (client_cmdline_parser(argc, argv, &ct->conf))
		goto out;
	HANDLE_VERSION_FLAG("client", ct->conf);

	ct->config_file = ct->conf.config_file_given?
		para_strdup(ct->conf.config_file_arg) :
		make_message("%s/.paraslash/client.conf", home);
	ret = file_exists(ct->config_file);
	if (!ret && ct->conf.config_file_given) {
		ret = -E_NO_CONFIG;
		goto out;
	}
	if (ret) {
		struct client_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 0
		};
		ret = -E_BAD_CONFIG;
		if (client_cmdline_parser_config_file(ct->config_file,
			&ct->conf, &params))
			goto out;
	}
	ct->user = ct->conf.user_given?
		para_strdup(ct->conf.user_arg) : para_logname();

	if (ct->conf.key_file_given)
		ct->key_file = para_strdup(ct->conf.key_file_arg);
	else {
		ct->key_file = make_message("%s/.paraslash/key.%s",
			home, ct->user);
		if (!file_exists(ct->key_file)) {
			free(ct->key_file);
			ct->key_file = make_message("%s/.ssh/id_rsa", home);
		}
	}

	if (loglevel)
		*loglevel = get_loglevel_by_name(ct->conf.loglevel_arg);
	PARA_INFO_LOG("loglevel: %s\n", ct->conf.loglevel_arg);
	PARA_INFO_LOG("config_file: %s\n", ct->config_file);
	PARA_INFO_LOG("key_file: %s\n", ct->key_file);
	ret = ct->conf.inputs_num;
out:
	free(home);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		client_close(ct);
		*ct_ptr = NULL;
	}
	return ret;
}

/**
 * Parse the client configuration and open a connection to para_server.
 *
 * \param argc See \ref client_parse_config.
 * \param argv See \ref client_parse_config.
 * \param ct_ptr See \ref client_parse_config.
 * \param loglevel See \ref client_parse_config.
 * \param parent See \ref client_connect().
 * \param child See \ref client_connect().
 * \param sched See \ref client_connect().
 *
 * This function combines client_parse_config() and client_connect(). It is
 * considered a syntax error if no command was given, i.e. if the number
 * of non-option arguments is zero.
 *
 * \return Standard.
 */
int client_open(int argc, char *argv[], struct client_task **ct_ptr,
		int *loglevel, struct btr_node *parent, struct btr_node *child,
		struct sched *sched)
{
	int ret = client_parse_config(argc, argv, ct_ptr, loglevel);

	if (ret < 0)
		return ret;
	if (ret == 0) {
		ret = -E_CLIENT_SYNTAX;
		goto fail;
	}
	ret = client_connect(*ct_ptr, sched, parent, child);
	if (ret < 0)
		goto fail;
	return 1;
fail:
	client_close(*ct_ptr);
	*ct_ptr = NULL;
	return ret;
}
