/*
 * Copyright (C) 2005-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file close_on_fork.c Manage a list of fds that should be closed on fork. */

#include <regex.h>

#include "para.h"
#include "list.h"
#include "string.h"

static struct list_head close_on_fork_list;
static int initialized;

/**
 * Describes an element of the close-on-fork list.
 *
 * \sa list.h
 */
struct close_on_fork {
	/** The file descriptor which should be closed after fork(). */
	int fd;
	/** The position in the close-on-fork list. */
	struct list_head node;
};

/**
 * Add one file descriptor to the close-on-fork list.
 *
 * \param fd The file descriptor to add.
 */
void add_close_on_fork_list(int fd)
{
	struct close_on_fork *cof = para_malloc(sizeof(struct close_on_fork));

	if (!initialized) {
		INIT_LIST_HEAD(&close_on_fork_list);
		initialized = 1;
	}
	cof->fd = fd;
	para_list_add(&cof->node, &close_on_fork_list);
}

/**
 * Delete one file descriptor from the close-on-fork list.
 *
 * \param fd The file descriptor to delete.
 *
 * Noop if \a fd does not belong to the close-on-fork list.
 */
void del_close_on_fork_list(int fd)
{
	struct close_on_fork *cof, *tmp;

	if (!initialized)
		return;
	list_for_each_entry_safe(cof, tmp, &close_on_fork_list, node) {
		if (fd != cof->fd)
			continue;
		list_del(&cof->node);
		free(cof);
	}
}

/**
 * Close all fds in the list and destroy all list entries.
 *
 * This function calls close(3) for each fd in the close-on-fork list
 * and empties the list afterwards.
 */
void close_listed_fds(void)
{
	struct close_on_fork *cof, *tmp;

	if (!initialized)
		return;
	list_for_each_entry_safe(cof, tmp, &close_on_fork_list, node) {
		PARA_DEBUG_LOG("closing fd %d\n", cof->fd);
		close(cof->fd);
		list_del(&cof->node);
		free(cof);
	}
}
