/*
 * Copyright (C) 2006-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_recv.c paraslash's dccp receiver */

/*
 * based on client.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "string.h"
#include "net.h"
#include "fd.h"
#include "buffer_tree.h"

#include "dccp_recv.cmdline.h"

static void dccp_recv_close(struct receiver_node *rn)
{
	if (rn->fd > 0)
		close(rn->fd);
	btr_pool_free(rn->btrp);
}

static int dccp_recv_open(struct receiver_node *rn)
{
	struct dccp_recv_args_info *conf = rn->conf;
	struct flowopts *fo = NULL;
	uint8_t *ccids = NULL;
	int fd, ret, i;

	/* Copy CCID preference list (u8 array required) */
	if (conf->ccid_given) {
		ccids = para_malloc(conf->ccid_given);
		fo    = flowopt_new();

		for (i = 0; i < conf->ccid_given; i++)
			ccids[i] = conf->ccid_arg[i];

		OPT_ADD(fo, SOL_DCCP, DCCP_SOCKOPT_CCID, ccids, i);
	}

	fd = makesock(IPPROTO_DCCP, 0, conf->host_arg, conf->port_arg, fo);
	free(ccids);
	if (fd < 0)
		return fd;
	/*
	 * Disable unused CCIDs: the receiver does not send any application
	 * data to the server. By shutting down this unused path we reduce
	 * internal processing costs, as the unused CCIDs (in the kernel) are
	 * then bypassed.
	 */
	if (shutdown(fd, SHUT_WR) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	rn->btrp = btr_pool_new("dccp_recv", 320 * 1024);
	rn->fd = fd;
	return 1;
err:
	close(fd);
	return ret;
}

/**
 * Check whether the host supports the requested 'ccid' arguments.
 * \param conf DCCP receiver arguments.
 * \return True if all CCIDs requested in \a conf are supported.
 */
static bool dccp_recv_ccid_support_check(struct dccp_recv_args_info *conf)
{
	uint8_t *ccids;
	int i, j, nccids;

	nccids = dccp_available_ccids(&ccids);
	if (nccids <= 0)
		return false;

	for (i = 0; i < conf->ccid_given; i++) {
		for (j = 0; j < nccids && ccids[j] != conf->ccid_arg[i]; j++)
			;
		if (j == nccids) {
			PARA_ERROR_LOG("'CCID-%d' not supported on this host.\n",
					conf->ccid_arg[i]);
			return false;
		}
	}
	return true;
}

static void *dccp_recv_parse_config(int argc, char **argv)
{
	struct dccp_recv_args_info *tmp = para_calloc(sizeof(*tmp));

	if (!dccp_recv_cmdline_parser(argc, argv, tmp) &&
	    dccp_recv_ccid_support_check(tmp))
		return tmp;
	free(tmp);
	return NULL;
}

static void dccp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);

	t->error = 0;
	if (generic_recv_pre_select(s, t) <= 0)
		return;
	para_fd_set(rn->fd, &s->rfds, &s->max_fileno);
}

static void dccp_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct btr_node *btrn = rn->btrn;
	struct iovec iov[2];
	int ret, iovcnt;
	size_t num_bytes;

	ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
	if (ret <= 0)
		goto out;
	iovcnt = btr_pool_get_buffers(rn->btrp, iov);
	ret = -E_DCCP_OVERRUN;
	if (iovcnt == 0)
		goto out;
	ret = readv_nonblock(rn->fd, iov, iovcnt, &s->rfds, &num_bytes);
	if (num_bytes == 0)
		goto out;
	if (num_bytes <= iov[0].iov_len) /* only the first buffer was filled */
		btr_add_output_pool(rn->btrp, num_bytes, btrn);
	else { /* both buffers contain data */
		btr_add_output_pool(rn->btrp, iov[0].iov_len, btrn);
		btr_add_output_pool(rn->btrp, num_bytes - iov[0].iov_len, btrn);
	}
out:
	if (ret >= 0)
		return;
	btr_remove_node(rn->btrn);
	t->error = ret;
}

static void dccp_recv_free_config(void *conf)
{
	dccp_recv_cmdline_parser_free(conf);
	free(conf);
}

/**
 * The init function of the dccp receiver.
 *
 * \param r Pointer to the receiver struct to initialize.
 *
 * Initialize all function pointers of \a r.
 */
void dccp_recv_init(struct receiver *r)
{
	struct dccp_recv_args_info dummy;

	dccp_recv_cmdline_parser_init(&dummy);
	r->open = dccp_recv_open;
	r->close = dccp_recv_close;
	r->pre_select = dccp_recv_pre_select;
	r->post_select = dccp_recv_post_select;
	r->parse_config = dccp_recv_parse_config;
	r->free_config = dccp_recv_free_config;
	r->help = (struct ggo_help) {
		.short_help = dccp_recv_args_info_help,
		.detailed_help = dccp_recv_args_info_detailed_help
	};
	dccp_recv_cmdline_parser_free(&dummy);
}
