/*
 * Copyright (C) 2006-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file recv_common.c common functions of para_recv and para_audiod */

#include <regex.h>
#include <stdbool.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "string.h"
#include "buffer_tree.h"

DEFINE_RECEIVER_ARRAY;

/**
 * Call the init function of each paraslash receiver.
 */
void recv_init(void)
{
	int i;

	FOR_EACH_RECEIVER(i)
		receivers[i].init(&receivers[i]);
}

static void *parse_receiver_args(int receiver_num, char *options)
{
	struct receiver *r = &receivers[receiver_num];
	char **argv;
	int argc, i;
	void *conf;


	PARA_DEBUG_LOG("%s, options: %s\n", r->name,
		options? options : "(none)");
	if (options) {
		PARA_DEBUG_LOG("options: %s\n", options);
		argc = create_argv(options, " \t", &argv);
		if (argc < 0)
			return NULL;
		for (i = argc - 1; i >= 0; i--)
			argv[i + 1] = argv[i];
		argc++;
	} else {
		argc = 1;
		argv = para_malloc(2 * sizeof(char*));
		argv[1] = NULL;
	}
	argv[0] = make_message("%s_recv", r->name);
	conf = r->parse_config(argc, argv);
	for (i = 0; i < argc; i++)
		free(argv[i]);
	free(argv);
	return conf;
}

/**
 * check if given string is a valid command line for any receiver
 *
 * \param \ra string of the form receiver_name:options
 * \param receiver_num contains the number of the receiver upon success
 *
 * This function checks whether \a ra starts with the name of a supported
 * paraslash receiver, optinally followed by a colon and any options for that
 * receiver. If a valid receiver name was found and further are present, the
 * remaining part of \a ra is passed to that receiver's config parser.
 *
 * \return On success, a pointer to the gengetopt args info struct is returned
 * and \a receiver_num contains the number of the receiver. Otherwise this function
 * returns \p NULL.
 */
void *check_receiver_arg(char *ra, int *receiver_num)
{
	int j;

	PARA_DEBUG_LOG("checking %s\n", ra);
	for (j = 0; receivers[j].name; j++) {
		const char *name = receivers[j].name;
		size_t len = strlen(name);
		char c;
		if (strlen(ra) < len)
			continue;
		if (strncmp(name, ra, len))
			continue;
		c = ra[len];
		if (c && c != ' ')
			continue;
		if (c && !receivers[j].parse_config)
			return NULL;
		*receiver_num = j;
		return parse_receiver_args(j, c? ra + len + 1: NULL);
	}
	PARA_ERROR_LOG("receiver not found\n");
	return NULL;
}

/**
 * Print out the help texts to all receivers.
 *
 * \param detailed Whether the detailed help should be printed.
 */
void print_receiver_helps(int detailed)
{
	int i;

	printf_or_die("\nAvailable receivers: \n\t");
	FOR_EACH_RECEIVER(i)
		printf_or_die("%s%s", i? " " : "", receivers[i].name);
	printf_or_die("\n\n");
	FOR_EACH_RECEIVER(i) {
		struct receiver *r = receivers + i;
		if (!r->help.short_help)
			continue;
		printf_or_die("Options for %s:\n", r->name);
		ggo_print_help(&r->help, detailed);
	}
}

/**
 * Simple pre-select hook, used by all receivers.
 *
 * \param s Scheduler info.
 * \param t Determines the receiver node.
 *
 * This requests a minimal delay from the scheduler if the status of the buffer
 * tree node indicates an error/eof condition. No file descriptors are added to
 * the fd sets of \a s.
 *
 * \return The status of the btr node of the receiver node, i.e. the return
 * value of the underlying call to \ref btr_node_status().
 */
int generic_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	int ret = btr_node_status(rn->btrn, 0, BTR_NT_ROOT);

	t->error = 0;
	if (ret < 0)
		sched_min_delay(s);
	return ret;
}
